unit xmlUtil;

// from https://mitjastachowiak.de/components/pascal/xmlUtil.pas

{$mode objfpc}{$H+}
{$modeswitch nestedprocvars}

interface

uses
  Classes, SysUtils, dialogs, LazUTF8;

type

 {
  TFastArray
  Eine Erweiterung für Arrays, welche die Funktion eines Sets hat. Die Positionen
  der Elemente sind hier nicht wichtig und können geändert werden. Der Objekt-Typ
  muss ein nilObj beinhalten, dass signalisiert, dass dieser Array-Eintrag leer ist.
  Beim iterieren über die Elemente müssen diese nil-Einträge übersprungen werden.
 }
 generic TFastArray<ObjType> = object
  type
   ArrayOfObjType = Array of ObjType;
  strict private
   emptyCnt : Cardinal;
   nilObj : ObjType;
   function readArr (i:Cardinal) : ObjType; inline;
   function readCount : Cardinal; inline;
  protected
   _arr : Array of ObjType;
   function cmp(const a, b : ObjType) : Boolean; inline;
  public
   constructor init(_nilObj : ObjType);
   procedure add(obj : ObjType);
   function remove(obj : ObjType) : Boolean;
   property arr[i:Cardinal] : ObjType read readArr; default;
   property count : Cardinal read readCount;
   destructor destroy;
 end;


 { XML }
 XML = class
  type
   TNodeType = (UNKNOWN_NODE = 0, ELEMENT_NODE = 1, TEXT_NODE = 3, COMMENT_NODE = 4);
   TElement = class;

   { TdtdInfo }
   TdtdInfo = packed record
    TagName  : shortString;
    // Legend: Optional, Forbidden, Empty, Deprecated, Loose DTD, Frameset DTD
    StartTag : Char;
    EndTag   : Char;
    Empty    : Char;
    Depr     : Char;
    DTD      : Char;
   end;
   TSymbol = packed record
    html : array [0..15] of Byte;
    utf8 : array [0..15] of Byte;
   end;

   { TNode }
   TNode = class (IFPObserved)
    strict private
     var
      observers : specialize TFastArray<TObject>;
     function readChanged : Boolean; inline;
     procedure writeBufferChangeEvents(v : Boolean);
    private
     _nodeType : TNodeType;
     _parentEl : TElement;
    protected
     _changed : Byte; // 1 = child nodes where changed   2 = element-properties were changed since load   4 = element was created since load   8 = a child existing since load was removed   16 = parentElement has changed since load
     _bufferChangeEvents : Boolean;
     destroying : Boolean;
     procedure setChanged(c : Byte); virtual;
     procedure FPONotifyObservers(ASender : TObject; AOperation : TFPObservedOperation; Data : Pointer);
    public
     constructor Create; reintroduce; virtual;
     property changed : Boolean read readChanged;
     procedure FPOAttachObserver(AObserver : TObject);
     procedure FPODetachObserver(AObserver : TObject);
     procedure markUnchanged; virtual;
     property bufferChangeEvents : Boolean read _bufferChangeEvents write writeBufferChangeEvents;
     property nodeType : TNodeType read _nodeType;
     property parentElement : TElement read _parentEl;
     function clone(res:TNode=nil) : TNode; virtual;
     destructor Destroy; override;
   end;

   { TAttribute }

   TAttribute = object
    public
     name  : ShortString;
     value : String;
   end;
   PAttribute = ^TAttribute;
   TAttributes = object (specialize TFastArray<TAttribute>)
   end;

   { TText }
   TText = class(TNode)
    type
     TTextChangeObs = class (IFPObserver)
      el : TText;
      procedure FPOObservedChanged({%H-}ASender : TObject; {%H-}Operation : TFPObservedOperation; {%H-}Data : Pointer);
     end;
    strict private
     textChangeObs : TTextChangeObs;
    private
     lastLineRemoved : Boolean;
    public
     text : TStrings;
     constructor Create (_text : TStrings = nil);
     function clone(res:TNode=nil) : TText; override;
     destructor Destroy; override;
   end;

   { TComment }
   TComment = class(TText)
    textBreak : Byte;   // 1 : Break before Begin   8 : Break after End
    constructor Create (_text : TStrings = nil); reintroduce;
    function clone(res:TNode=nil) : TComment; override;
   end;

   { TCData }
   TCData = class(TComment)
    data : Array of Byte;
    constructor Create; reintroduce;
   end;

   { TElement }
   TElement = class(TNode)
    type
     ArrayOfTElement = Array of TElement;
     TElementIterator = procedure (el:TElement) is nested;
     TAttributeIterator = procedure (attribute:ShortString; value:String) is nested;
     TChildNodes = object
      private
       n : Array of TNode;
       function ReadNodes(i : Cardinal) : TNode; inline;
       function ReadCount : Cardinal; inline;
      public
       property nodes[i : Cardinal] : TNode read ReadNodes; default;
       property count : Cardinal read ReadCount;
     end;
     PChildNodes = ^TChildNodes;
    private
     shift         : integer;  // Merker für Shift bei ToString
     __tagName     : ShortString;
     __selfClosing : Boolean;
     procedure setTagName(v : ShortString);
     function getElement(s : String; i : integer) : TElement;
    protected
     attributes : TAttributes;
     procedure setChanged(c : Byte); override;
    public
     textBreak  : Byte;   // 1 : Break before Begin   2 : Break after Begin    4 : Break before End   8 : Break after End
     childNodes : TChildNodes;
     constructor Create; overload; override;
     constructor Create(_tagName : ShortString); overload;
     function clone(res:TNode=nil) : TElement; override;
     property tagName : ShortString read __tagName write setTagName;
     property selfClosing : Boolean read __selfClosing;
     procedure appendChild(newChild:TNode);
     procedure insertChildBefore(newChild:TNode; predecessor:TNode);
     procedure markUnchanged; override;
     procedure removeChild(i : Cardinal); overload;
     procedure removeChild(child : TNode); overload;
     function getAttribute(name : ShortString) : String;
     procedure setAttribute(name : ShortString; value : String);
     function getElementsByTagName(_tagName : ShortString; i : Cardinal) : TElement; overload;
     function getElementsByTagName(_tagName : ShortString) : ArrayOfTElement; overload;
     function getElementById(id : String) : TElement;
     function getTextContent(RemoveSpaces : Boolean) : TStrings; overload;
     function getTextContent : String; overload;
     function getSingleLineContent : String; overload;
     function firstChild : TNode;
     procedure iterateChildElements(it:TElementIterator); // to call this method, use {$modeswitch nestedprocvars} in your unit!
     procedure iterateAttributes(it:TAttributeIterator);  // to call this method, use {$modeswitch nestedprocvars} in your unit!
     procedure setTextContent(s : String);
     procedure clearNodes;
     destructor destroy; override;
   end;

   { TDocument }

   TDocument = class(TElement)
    public
     constructor Create; reintroduce;
     class function fromStrings(str:TStrings) : XML.TDocument;
     class function fromString(str:string) : XML.TDocument;
     function toStrings : TStrings;
   end;

 private
  class var dtd         : Array of TdtdInfo;
  class var symbolTable : Array of TSymbol;
 public
  class procedure SetSymbolTable(str : TStrings);
  class procedure SetDTD(str : TStrings);
  class function HTMLToUTF8(const h : String) : String;
  class function UTF8ToHTML(const s : String; escapeLineBreaks : Boolean = false) : String;
 end;

implementation

operator = (a, b : XML.TAttribute) r : boolean; // compares the name of two attributes (not the values!)
begin
 r := a.name = b.name;
end;

{ XML.TCData }

constructor XML.TCData.Create;
begin
  inherited Create;
  data := nil;
end;



constructor TFastArray.init(_nilObj : ObjType);
begin
 _arr := nil;
 emptyCnt := 0;
 nilObj := _nilObj;
end;

function TFastArray.readCount : Cardinal;
begin
 Result := Length(_arr);
end;

function TFastArray.cmp(const a, b: ObjType): Boolean;
begin
 Result := a = b;
end;

function TFastArray.readArr(i : Cardinal) : ObjType;
begin
 Result := _arr[i];
end;

procedure TFastArray.add(obj : ObjType);
var i : integer;
begin
 if (emptyCnt = 0) then begin
  i := Length(_arr);
  setLength(_arr, i+(i div 3)+1);
  _arr[i] := obj;
  emptyCnt := Length(_arr) - i - 1;
  for i := i + 1 to Length(_arr)-1 do _arr[i] := nilObj;
 end else for i := Length(_arr)-1 downto 0 do if (_arr[i] = nilObj) then begin
  _arr[i] := obj;
  emptyCnt := emptyCnt-1;
  break;
 end;
end;

function TFastArray.remove(obj : ObjType) : Boolean;
var i,j : integer;
begin
 Result := false;
 if (emptyCnt {%H-}>= Cardinal(Length(_arr) div 2 + 1)) then j := 0 else j := -1;
 for i := 0 to Length(_arr)-1 do begin
  if (_arr[i] = obj) then begin
   Result := true;
   _arr[i] := nilObj;
   if (j = -1) then begin emptyCnt := emptyCnt+1; exit; end;
  end;
  if (_arr[i] <> nilObj) and (j <> -1) then begin
   if (i <> j) then begin
    _arr[j] := _arr[i];
    _arr[i] := nilObj;
   end;
   j := j + 1;
  end;
 end;
 if (j = -1) then exit; // not found
 SetLength(_arr, j + j div 3);
 emptyCnt := Length(_arr) - j;
end;

destructor TFastArray.destroy;
begin
 SetLength(_arr, 0);
end;




class procedure XML.SetSymbolTable(str : TStrings);
var
  i, j, k : integer;
  s : TSymbol;
begin
 setLength(symbolTable, str.Count);
 for i := 0 to str.count - 1 do begin
  j := 1;
  while (j < Length(str[i])) and (Byte(str[i][j]) >= 128) do j := j + 1;
  for k := 0 to 15 do begin s.html[k] := Byte(str[i][j]); j := j + 1; if (s.html[k] = 9) or (s.html[k] = 0) then begin s.html[k] := 0; break; end; end;
  for k := 0 to 14 do begin s.utf8[k] := Byte(str[i][j]); if (j = Length(str[i])) then begin s.utf8[k+1] := 0; break; end; j := j + 1; end;
  symbolTable[i] := s;
 end;
end;

// returns the _value with html-entities converted to UTF-8
class function XML.HTMLToUTF8(const h : String) : String;
var
  d    : smallint;
  c    : cardinal;
  i, i_, j, m : integer;
const
  UNICODE_MAX = $10FFFF;
  entityNames : Array [0..194] of Array [0..11] of Char = ('aacute','acirc','acute','add','aelig','agrave','alefsym','alpha','amp','and','ang','apos','aring','asymp','atilde','auml','bdquo','beta','brvbar','bull','cap','ccedil','cedil','cent','chi','circ','clubs','cong','copy','crarr','cup','curren','dagger','darr','deg','delta','diams','divide','eacute','ecirc','egrave','empty','emsp','ensp','epsilon','equal','equiv','eta','eth','euml','euro','exclamation','exist','fnof','forall','frac12','frac14','frac34','frasl','gamma','ge','gt','harr','hearts','hellip','horbar','iacute','icirc','iexcl','igrave','image','infin','int','iota','iquest','isin','iuml','kappa','lambda','lang','laquo','larr','lceil','ldquo','le','lfloor','lowast','loz','lrm','lsaquo','lsquo','lt','macr','mdash','micro','middot','minus','mu','nabla','nbsp','ndash','ne','ni','not','notin','nsub','ntilde','nu','oacute','ocirc','oelig','ograve','oline','omega','omicron','oplus','or','ordf','ordm','oslash','otilde','otimes','ouml','para','part','percent','permil','perp','phi','pi','piv','plusmn','pound','prime','prod','prop','psi','quot','radic','rang','raquo','rarr','rceil','rdquo','real','reg','rfloor','rho','rlm','rsaquo','rsquo','sbquo','scaron','sdot','sect','shy','sigma','sigmaf','sim','spades','sub','sube','sum','sup','sup1','sup2','sup3','supe','szlig','tau','there4','theta','thetasym','thinsp','thorn','tilde','times','trade','uacute','uarr','ucirc','ugrave','uml','upsih','upsilon','uuml','weierp','xi','yacute','yen','yuml','zeta','zwj','zwnj','');
  entityCodeL : Array [0..194] of Word = (225,226,180,43,230,224,8501,945,38,8743,8736,39,229,8776,227,228,8222,946,166,8226,8745,231,184,162,967,710,9827,8773,169,8629,8746,164,8224,8595,176,948,9830,247,233,234,232,8709,8195,8194,949,61,8801,951,240,235,8364,33,8707,402,8704,189,188,190,8260,947,8805,62,8596,9829,8230,8213,237,238,161,236,8465,8734,8747,953,191,8712,239,954,955,9001,171,8592,8968,8220,8804,8970,8727,9674,8206,8249,8216,60,175,8212,181,183,8722,956,8711,160,8211,8800,8715,172,8713,8836,241,957,243,244,339,242,8254,969,959,8853,8744,170,186,248,245,8855,246,182,8706,37,8240,8869,966,960,982,177,163,8242,8719,8733,968,34,8730,9002,187,8594,8969,8221,8476,174,8971,961,8207,8250,8217,8218,353,8901,167,173,963,962,8764,9824,8834,8838,8721,8835,185,178,179,8839,223,964,8756,952,977,8201,254,732,215,8482,250,8593,251,249,168,978,965,252,8472,958,253,165,255,950,8205,8204,0);
  entityCodeU : Array [0..194] of Word = (193,194,0,0,198,192,0,913,0,0,0,0,197,0,195,196,0,914,0,0,0,199,0,0,935,0,0,0,0,0,0,0,8225,8659,0,916,0,0,201,202,200,0,0,0,917,0,0,919,208,203,0,0,0,0,0,0,0,0,0,915,0,0,8660,0,0,0,205,206,0,204,0,0,0,921,0,0,207,922,923,0,0,8656,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,924,0,0,0,0,0,0,0,0,209,925,211,212,338,210,0,937,927,0,0,0,0,216,213,0,214,0,0,0,0,0,934,928,0,0,0,8243,0,0,936,0,0,0,0,8658,0,0,0,0,0,929,0,0,0,0,352,0,0,0,931,0,0,0,0,0,0,0,0,0,0,0,0,932,0,920,0,0,222,0,0,0,218,8657,219,217,0,0,933,220,0,926,221,0,376,918,0,0,0);
begin
 Result := h.Copy(h);
 j := 1;
 i := 1;
 while (i <= Length(h)) do begin
  if (i <= Length(h) - 4) and (h[i] = '&') then begin // detect HTML entities
   i_ := i;
   c := 0;
   d := 255; // 255 or 254 means error
   if (h[i+1] = '#') and (h[i+2] = 'x') then begin // read hex number value c
    i := i + 3;
    while (i < Length(h)) do begin
     d := (Byte(h[i]) or 32) - 48;
     if (d > 9) then d := d - 39;
     if (d < 0) or (d >= 16) then begin
      if (h[i] <> ';') then d := 255;
      break;
     end;
     c := c*16 + d;
     if (c > UNICODE_MAX) then begin d := 255; break; end;
     i := i + 1;
    end;
   end else if (h[i+1] = '#') then begin // read dec number value c
    i := i + 2;
    while (i < Length(h)) do begin
     d := Byte(h[i]) - 48;
     if (d > 9) then begin
      if (h[i] <> ';') then d := 255;
      break;
     end;
     c := c*10 + d;
     if (c > UNICODE_MAX) then begin d := 255; break; end;
     i := i + 1;
    end;
   end else begin // determine char via table
    i := i + 1;
    m := 0;
    while (i <= Length(h)) do begin
     if (h[i] = ';') then begin
      if (d = 254) then c := entityCodeU[m]
      else c := entityCodeL[m];
      if (c <> 0) then d := 0; // entity found!
      break;
     end;
     if (c = 0) then begin // select start search index according to alphabet linear mapping
      m := ((Byte(h[i]) or 32) - 97) * (Length(entityNames)-1) div 25;
      if (m < 0) or (m >= Length(entityNames)) then break; // unerlaubtes Zeichen
     end;
     while (Byte(entityNames[m][c]) > Byte(h[i]) or 32) and (m > 0) and ((c = 0) or (Byte(entityNames[m][c-1]) = Byte(entityNames[m-1][c-1]))) do m := m - 1;
     while (Byte(entityNames[m][c]) < Byte(h[i]) or 32) and (m < Length(entityNames)-1) and ((c = 0) or (Byte(entityNames[m][c-1]) = Byte(entityNames[m+1][c-1]))) do m := m + 1;
     if (Byte(entityNames[m][c]) <> Byte(h[i]) or 32) then break;
     if (Byte(h[i]) and 32 = 0) then d := 254; // indicate, that there is at least one upper case character
     i := i + 1;
     c := c + 1;
    end;
   end;
   if (d >= 254) then begin  // error: HTML entity incorrect or unknown
    i := i_;
    Result[j] := '&';
   end else j := j - 1                                 // -1, because normal chars are also 1 byte long
                       + UnicodeToUTF8(c, @Result[j]); // convert s to UTF-Codepoint and copy the utf-char(s) into the result (the UTF-8 representation will always be shorter (max. 4 bytes), than the HTML-entity (min. 4 bytes))
  end else if (j <> i) then Result[j] := h[i];
  i := i + 1;
  j := j + 1;
 end;
 SetLength(Result, j-1);
end;

class function XML.UTF8ToHTML(const s : String; escapeLineBreaks : Boolean = false) : String;
var
  p       : PChar;
  unicode : Cardinal;
  CharLen : integer;
begin
 Result := '';
 p := PChar(s);
 repeat
  unicode := UTF8CharacterToUnicode(p, CharLen);
  if (unicode = 0) then break;
  if (unicode <= 126) and (unicode >= 32)
    and (unicode <> Byte('<'))
    and (unicode <> Byte('>'))
    and (unicode <> Byte('&'))
    and (unicode <> Byte('"'))
    or (not escapeLineBreaks) and ((unicode = Byte(13)) or (unicode = Byte(10))) then Result := Result + Char(unicode)
  else Result := Result + '&#' + IntToStr(unicode) + ';';
  inc(p, CharLen);
 until (CharLen = 0);
end;

class procedure XML.SetDTD(str : TStrings);
var
 i, j, k, l : integer;
 d          : TdtdInfo;
 s          : ShortString;
procedure QuickSortDTD(var A : Array of XML.TdtdInfo; _lo, _hi : integer);
var
 lo,hi : integer;
 m,s : XML.TdtdInfo;
begin
 m := A[(_lo + _hi) div 2];
 lo := _lo;
 hi := _hi;
 repeat
  while (CompareStr(A[lo].TagName, m.TagName) < 0) do lo := lo+1;
  while (CompareStr(A[hi].TagName, m.TagName) > 0) do hi := hi-1;
  if (lo <= hi) then begin
   s := A[lo];
   A[lo] := A[hi];
   A[hi] := s;
   lo := lo+1;
   hi := hi-1;
  end;
 until (lo > hi);
 if (hi > _lo) then QuickSortDTD(A,_lo,hi);
 if (lo < _hi) then QuickSortDTD(A,lo,_hi);
end;
begin
 SetLength(dtd, str.Count);  // Ausreichend Platz in dtd sicherstellen
 j := 0;
 for i := 0 to str.Count-1 do begin
  // Split Line into d
  FillByte(d{%H-}, SizeOf(d), 0);
  l := 0;
  s := '';
  for k := 1 to Length(str[i]) do if (str[i][k] = Char(9)) or (k = Length(str[i])) then begin
   case l of
    0 : d.TagName := UpperCase(s);
    1 : d.StartTag := s[1];
    2 : d.EndTag := s[1];
    3 : d.Empty := s[i];
    4 : d.Depr := s[1];
    5 : d.DTD := s[1];
   end;
   l := l + 1;
   s := '';
  end else s := s + str[i][k];
  // Falls d gültig, in dtd speichern
  if (Length(d.TagName) <> 0) then begin
   dtd[j] := d;
   j := j + 1;
  end;
 end;
 SetLength(dtd, j); // leeren Abschnitt von dtd löschen
 // dtd sortieren
 if (j = 0) then dtd := nil
 else QuickSortDTD(dtd, 0, j-1);
end;



constructor XML.TText.Create(_text : TStrings = nil);
begin
 inherited Create;
 lastLineRemoved := false;
 _nodeType := TEXT_NODE;
 _parentEl := nil;
 text := _text;
 if (text = nil) then text := TStringList.Create;
 setChanged(_changed or 4);
 textChangeObs := XML.TText.TTextChangeObs.Create;
 textChangeObs.el := self;
 text.FPOAttachObserver(textChangeObs);
end;

function XML.TText.clone(res:TNode=nil): XML.TText;
begin
 if (res = nil) then res := TText.Create;
 Result := XML.TText(inherited clone(res));
 Result.text.Assign(self.text);
 Result._changed := _changed;
end;

procedure XML.TText.TTextChangeObs.FPOObservedChanged(ASender : TObject; Operation : TFPObservedOperation; Data : Pointer);
begin
 el.SetChanged(_changed or 2);
end;

destructor XML.TText.Destroy;
begin
 text.Free;
 textChangeObs.Free;
 inherited;
end;




constructor XML.TComment.Create(_text : TStrings = nil);
begin
 inherited Create(_text);
 _nodeType := COMMENT_NODE;
 textBreak := 0;
end;

function XML.TComment.clone(res: TNode=nil): TComment;
begin
 if (res = nil) then res := XML.TComment.Create;
 Result := XML.TComment(inherited clone(res));
 Result.textBreak := self.textBreak;
end;




constructor XML.TElement.Create;
var nilAtr : TAttribute;
begin
 inherited;
 _nodeType := ELEMENT_NODE;
 _parentEl := nil;
 __selfClosing := false;
 __tagName := '';
 textBreak := 0;
 SetLength(childNodes.n, 0);
 nilAtr.name := '';
 nilAtr.value := '';
 attributes.init(nilAtr);//SetLength(attributes.a, 0);
 setChanged(_changed or 4);
end;

constructor XML.TElement.Create(_tagName : ShortString);
begin
 Create;
 setTagName(_tagName);
end;

function XML.TElement.clone(res: TNode): TElement;
var i : integer;
begin
 if (res = nil) then res := TElement.Create(self.tagName);
 Result := XML.TElement(inherited clone(res));
 for i := 0 to self.childNodes.count - 1 do Result.appendChild(self.childNodes[i].clone);
 for i := 0 to self.attributes.count-1 do if (self.attributes.arr[i].name <> '') then Result.setAttribute(self.attributes.arr[i].name, self.attributes.arr[i].value);
 Result._changed := self._changed;
end;

procedure XML.TElement.setTagName(v : ShortString);
var u, o, m : integer;
begin
 __tagName := UpperCase(v);
 __selfClosing := false;
 if (dtd <> nil) then begin
  // TagName in dtd suchen
  u := 0;
  o := Length(dtd)-1;
  repeat begin
   m := (u + o) div 2;
   if (CompareStr(dtd[m].TagName, __tagName) >= 0) then o := m else u := m+1;
  end until (u = o);
  // SelfClosing auslesen
  if (dtd[o].TagName = __tagName) then __selfClosing := dtd[o].EndTag = 'F';
 end;
end;

function XML.TElement.TChildNodes.ReadNodes(i : Cardinal) : TNode; inline;
begin
 Result := n[i];
end;

function XML.TElement.TChildNodes.ReadCount : Cardinal; inline;
begin
 Result := Length(n);
end;

procedure XML.TElement.appendChild(newChild: TNode);
begin
 insertChildBefore(newChild, nil);
end;

procedure XML.TElement.insertChildBefore(newChild:TNode; predecessor:TNode);
var i,j : integer;
begin
 if (NewChild._parentEl <> nil) then begin
  NewChild._parentEl.removeChild(NewChild);
  NewChild.setChanged(newChild._changed or 16);
 end;
 i := 0;
 while (i < childNodes.count) do begin
  if (childNodes[i] = predecessor) then break;
  i += 1;
 end;
 j := Length(childNodes.n);
 SetLength(childNodes.n, j+1);
 for j := j downto i+1 do childNodes.n[j] := childNodes.n[j-1];
 childNodes.n[i] := newChild;
 NewChild._parentEl := self;
 if (NewChild._changed <> 0) then setChanged(_changed or 1);
end;

procedure XML.TElement.removeChild(i : Cardinal); overload;
var b : Boolean;
begin
 if (i >= childNodes.Count) then exit;
 b := childNodes.n[i]._changed and 4 <> 0;
 for i := i + 1 to Length(childNodes.n) - 1 do childNodes.n[i-1] := childNodes.n[i];
 SetLength(childNodes.n, Length(childNodes.n)-1);
 if (b) then setChanged(_changed and not 1)
 else setChanged(_changed or 8);
end;

procedure XML.TElement.removeChild(child : TNode); overload;
var
  i : integer;
begin
 for i := Length(childNodes.n)-1 downto 0 do if (childNodes.n[i] = child) then begin
  removeChild(i);
  break;
 end;
end;

function XML.TElement.getAttribute(name : ShortString) : String;
var i : integer;
begin
 Result := '';
 for i := 0 to attributes.count-1 do if (attributes[i].name = name) then begin
  Result := attributes[i].value;
  break;
 end;
end;

procedure XML.TElement.setAttribute(name : ShortString; value : String);
var
 i : integer;
 atr : TAttribute;
begin
 if (name = '') then exit;
 atr.name := name;
 atr.value := value;
 if (value = '') then begin
  attributes.remove(atr);
  exit;
 end;
 for i := 0 to attributes.count-1 do if (attributes[i].name = name) then begin
  if (attributes._arr[i].value <> value) then begin
   attributes._arr[i].value := value;
   setChanged(_changed or 2);
  end;
  exit;
 end;
 attributes.add(atr);
 setChanged(_changed or 2);
end;

function XML.TElement.getElement(s : String; i : integer) : TElement;
var
 j : integer;
procedure scan(el : TElement);
var k : integer;
begin
 if (el <> self) and (i < 0) and (el.getAttribute('id') = s) then begin Result := el; exit; end;
 if (el <> self) and (i >= 0) and (el.tagName = s) then begin
  j := j + 1;
  if (j = i) then begin Result := el; exit; end;
 end;
 for k := 0 to Length(el.childNodes.n) - 1 do begin
  if (el.childNodes.n[k].nodeType = ELEMENT_NODE) then scan(TElement(el.childNodes.n[k]));
  if (Result <> nil) then break;
 end;
end;
begin
 j := -1;
 Result := nil;
 scan(self);
end;

function XML.TElement.getElementsByTagName(_tagName : ShortString; i : Cardinal) : TElement;
begin
 Result := getElement(_tagName, i);
end;

function XML.TElement.getElementsByTagName(_tagName : ShortString) : ArrayOfTElement;
var l : integer;
procedure scan(el : TElement);
var i : integer;
begin
 for i := 0 to el.childNodes.count-1 do if (el.childNodes[i].nodeType = ELEMENT_NODE) then begin
  if (TElement(el.childNodes[i]).tagName = _tagName) then begin
   if (Length(Result) <= l) then setLength(result, Length(Result)*2);
   result[l] := TElement(el.childNodes[i]);
   l := l + 1;
  end;
  scan(TElement(el.childNodes[i]));
 end;
end;
begin
 Result := nil;
 l := 0;
 SetLength(Result, 4);
 scan(self);
 SetLength(Result, l);
end;

function XML.TElement.getElementById(id : String) : TElement;
begin
 Result := getElement(id, -1);
end;

function XML.TElement.getTextContent(RemoveSpaces : Boolean) : TStrings;
function chomp (const s : String) : String;
var
  start, stop : integer;
label SearchStart;
label SearchStop;
begin
 Result := '';
 stop := Length(s);
 SearchStop:
 if (stop <= 1) then exit;
 case Byte(s[stop]) of
  13 : begin stop := stop - 1; goto SearchStop; end;
  10 : begin stop := stop - 1; goto SearchStop; end;
  32 : begin stop := stop - 1; goto SearchStop; end;
 end;
 start := 1;
 SearchStart:
 case Byte(s[start]) of
  13 : begin start := start + 1; goto SearchStart; end;
  10 : begin start := start + 1; goto SearchStart; end;
  32 : begin start := start + 1; goto SearchStart; end;
 end;
 SetLength(Result, stop - start + 1);
 move(s[start],Result[1],stop-start+1);
end;
procedure append (el : TElement);
var
  s    : String;
  i, j : integer;
begin
 for i := 0 to Length(el.childNodes.n) - 1 do case el.childNodes[i].nodeType of
  ELEMENT_NODE : append(XML.TElement(el.childNodes[i]));
  TEXT_NODE : begin
   for j := 0 to XML.TText(el.childNodes[i]).text.Count - 1 do if (RemoveSpaces) then begin
    s := chomp(XML.TText(el.childNodes[i]).text[j]);
    if (s <> '') then Result.add(s);
   end else Result.add(XML.TText(el.childNodes[i]).text[j]);
  end;
 end;
end;
begin
 Result := TStringList.Create;
 append(self);
end;

function XML.TElement.getTextContent : String;
var
  s    : TStrings;
begin
 s := getTextContent(true);
 Result := s.Text;
 s.Free;
end;

function XML.TElement.getSingleLineContent : String;
var
  s    : TStrings;
  i, j : integer;
  f    : Boolean;
begin
 s := getTextContent(true);
 Result := '';
 f := false;
 for i := 0 to s.Count - 1 do begin
  for j := 1 to Length(s[i]) do if (Byte(s[i][j]) <> 13) and (Byte(s[i][j]) <> 10) and ((f) or (Byte(s[i][j]) <> 32)) then begin
   Result := Result + s[i][j];
   f := true;
  end;
  if (f) then break;
 end;
 s.Free;
end;

function XML.TElement.firstChild: TNode;
begin
  if (childNodes.count > 0) then Result := childNodes[0]
  else Result := nil;
end;

procedure XML.TElement.iterateChildElements(it: TElementIterator);
var i : integer;
begin
  for i := 0 to Length(childNodes.n)-1 do if (childNodes.n[i].nodeType = ELEMENT_NODE) then begin
    it(TElement(childNodes.n[i]));
    TElement(childNodes.n[i]).iterateChildElements(it);
  end;
end;

procedure XML.TElement.iterateAttributes(it: TAttributeIterator);
var i : integer;
begin
  for i := 0 to attributes.count-1 do if (attributes.arr[i].name <> '') then it(attributes.arr[i].name, attributes.arr[i].value);
end;

procedure XML.TElement.setTextContent(s : String);
var
  i   : integer;
  t   : TText;
  str : TStringList;
begin
  clearNodes;
  str := TStringList.Create;
  if (str.Count = 0) then str.add('');
  for i := 1 to Length(s) do
    if (Byte(s[i]) = 0) then break
    else if (s[i] = '<') or (s[i] = '>') or (Byte(s[i]) >= 128) then begin
      ShowMessage('Illegal Character for TextContent: #' + IntToStr(Byte(s[i])) + '!');
      str.free;
      exit;
    end
    else if (s[i] = #10) then str.add('')
    else if (s[i] <> #13) then str[str.Count-1] := str[str.Count-1] + s[i];
  t := TText.Create(str);
  appendChild(t);
end;

procedure XML.TElement.clearNodes;
var i : integer;
begin
 for i := Length(childNodes.n) - 1 downto 0 do begin
  childNodes.n[i]._parentEl := nil;
  if (childNodes.n[i]._changed and 4 = 0) and (childNodes.n[i].nodeType <> UNKNOWN_NODE) then setChanged(_changed or 8);
  childNodes.n[i].destroy;
 end;
 SetLength(childNodes.n, 0);
 setChanged(_changed and not 1);
end;

procedure XML.TElement.markUnchanged;
var i : integer;
begin
 _changed := 0;
 for i := 0 to childNodes.count-1 do childNodes[i].markUnchanged;
 FPONotifyObservers(self, ooChange, nil);
 inherited;
end;

procedure XML.TElement.setChanged(c : Byte);
var i : integer;
begin
 if (_changed and 1 <> 0) and (c and 1 = 0) then for i := 0 to childNodes.count-1 do if (childNodes[i]._changed <> 0) then begin
  c := c or 1;
  break;
 end;
 inherited;
end;

destructor XML.TElement.destroy;
begin
 destroying := true;
 self.clearNodes;
 attributes.destroy;
 inherited;
end;



constructor XML.TNode.Create;
begin
 inherited;
 _nodeType := UNKNOWN_NODE;
 _changed := 0;
 _bufferChangeEvents := false;
 destroying := false;
 observers.init(nil);
end;

procedure XML.TNode.FPOAttachObserver(AObserver: TObject);
begin
 observers.add(AObserver);
end;

procedure XML.TNode.FPODetachObserver(AObserver: TObject);
begin
 observers.remove(AObserver);
end;

procedure XML.TNode.FPONotifyObservers(ASender: TObject; AOperation: TFPObservedOperation; Data: Pointer);
var
 i : integer;
begin
 if (destroying) and (AOperation <> ooFree) then exit;
 for i := 0 to observers.count-1 do if (observers[i] <> nil) then (observers[i] as IFPObserver).FPOObservedChanged(ASender, AOperation, Data);
end;

function XML.TNode.readChanged : Boolean;
begin
 Result := _changed <> 0;
end;

procedure XML.TNode.writeBufferChangeEvents(v: Boolean);
begin
 _bufferChangeEvents := v;
 if (not v) then SetChanged(_changed);
end;

procedure XML.TNode.markUnchanged;
begin
 _changed := 0;
 if (self.parentElement <> nil) then self.parentElement.setChanged(self.parentElement._changed and not 1);
end;

function XML.TNode.clone(res: TNode): TNode;
begin
 if (res = nil) then Result := TNode.Create else Result := res;
 Result._changed := self._changed;
end;

procedure XML.TNode.setChanged(c : Byte);
begin
 _changed := c;
 if (_bufferChangeEvents) then exit;
 FPONotifyObservers(self, ooChange, nil);
 if (_changed <> 0) and (self.parentElement <> nil) then self.parentElement.setChanged(self.parentElement._changed or 1);
 if (_changed = 0) and (self.parentElement <> nil) then self.parentElement.setChanged(self.parentElement._changed and not 1); // TElement.setChanged checks weather no other child node is changed before applying the 0
end;

destructor XML.TNode.Destroy;
begin
 if (_parentEl <> nil) then _parentEl.removeChild(self);
 FPONotifyObservers(self, ooFree, nil);
 observers.destroy;
 inherited;
end;



constructor XML.TDocument.Create;
begin
 inherited Create('');
end;

class function XML.TDocument.fromStrings(str: TStrings) : XML.TDocument;
begin
  Result := fromString(str.Text);
end;

class function XML.TDocument.fromString(str: string) : XML.TDocument;
var
  row, column, pos : integer;
  k, l : integer;
  el, pel : XML.TElement;
  tx : XML.TText;
  co : XML.TComment;
  pn : Array of TNode;
  doubleBreak : Boolean;
  c : Char;   // current char
  s : ShortString;
  t : String;
label init, textNode, xmlBegin, xmlTagName, attributeName, attributeValue, xmlEnd, commentNode, binaryBlock, endOfFile;
procedure next; inline;
label nextStart;
begin
  nextStart:
  if (pos > length(str)) then c := #0
  else c := str[pos];
  pos += 1;
  column += 1;
  if (c = #13) then goto nextStart;
  if (c = #10) then begin
    column := 1;
    row += 1;
  end;
end;
begin
  Result := XML.TDocument.Create;
  pos := 1;
  row := 1;
  column := 1;
  el := Result;
  tx := nil;
  co := nil;
  doubleBreak := false;
  Next;
  try
    init: // kein Status, von hier wird entweder nach textNode oder nach xmlBegin gesprungen
    if (c = '<') then begin
      next;
      goto xmlBegin;
    end else if (c <> ' ') and (c <> #9) and (c <> #10) then begin
      tx := TText.Create(nil);
      if (tx.text.Count = 0) then tx.text.add('');
      el.appendChild(tx);
      goto textNode;
    end else if (c = #0) then goto endOfFile;
    if (c = #10) then begin
      pn := el.childNodes.n;
      k := length(pn)-1;
      if (k = -1) then el.textBreak := el.textBreak or 2
      else if (pn[k].nodeType = ELEMENT_NODE) then begin
        if (TElement(pn[k]).textBreak and 8 = 8) then doubleBreak := true;
        TElement(pn[k]).textBreak := TElement(pn[k]).textBreak or 8;
      end else if (pn[k].nodeType = COMMENT_NODE) then begin
        if (TComment(pn[k]).textBreak and 8 = 8) then doubleBreak := true;
        TComment(pn[k]).textBreak := TComment(pn[k]).textBreak or 8;
      end;
    end;
    next;
    goto init;

    textNode: // innerhalb einer Text-Node, springt nach init zurück
    if (c = '<') then begin
      k := tx.text.Count-1;
      tx.LastLineRemoved := true;
      for l := Length(tx.text[k]) downto 1 do
        if (tx.text[k][l] <> ' ') and (tx.text[k][l] <> Char(9)) then begin
          tx.LastLineRemoved := false;
          break;
        end;
      if (tx.LastLineRemoved) then tx.text.delete(k);
      goto init;
    end else if (c = #0) then goto endOfFile;
    k := tx.text.Count-1;
    if (c = #10) then tx.text.add('')
    else tx.text[k] := tx.text[k] + Char(c);
    next;
    goto textNode;

    xmlBegin: // innerhalb eines XML-Begins, erstellt entweder ein neues Element und springt nach 3 oder springt nach 6
    if (c = ' ') or (c = #10) or (c = #9) then begin
      next;
      goto xmlBegin;
    end else if (c = '!') then begin
      if (pos < length(str) - 8) and (copy(str, pos, 7) = 'BINDATA') then begin // Element ist ein Datenblock
        pos += 7;
        row := Integer.minValue;
        goto binaryBlock;
      end else begin // Element ist ein Comment
        co := TComment.Create(nil);
        if (co.text.Count = 0) then co.text.add('');
        if (doubleBreak) then co.textBreak := co.textBreak or 1;
        k := Length(Result.childNodes.n);
        el.appendChild(co);
        next;
        goto commentNode;
      end;
    end else if (c = '/') then begin
      next;
      s := '';
      if (tx <> nil) and (tx.LastLineRemoved) then el.textBreak := el.textBreak or 4;
      goto xmlEnd;
    end else if (c = '>') then begin
      next;
      goto init;
    end else if (c = #0) then raise Exception.create('Unexpected end of file!')
    else begin
      s := '';
      pel := el;
      el := XML.TElement.Create;
      if (tx <> nil) and (tx.lastLineRemoved) then el.textBreak := el.textBreak or 1;
      if (pel = nil) then begin
        k := Length(Result.childNodes.n);
        SetLength(Result.childNodes.n, k+1);
        Result.childNodes.n[k] := el;
      end else begin
        if (doubleBreak) then el.textBreak := el.textBreak or 1;
        pel.appendChild(el);
      end;
      goto xmlTagName;
    end;

    xmlTagName: // TagName lesen, springt nach init oder nach attributeName
    if (c = ' ') or (c = #10) or (c = #9) or (c = '>') then begin
      el.tagName := s;
      s := '';
      if (c = '>') then begin
        if (el.selfClosing) then begin
          el := el.parentElement;
          if (el = nil) then raise Exception.Create('More elements where closed than created!');
        end;
        next;
        goto init;
      end else goto attributeName;
    end else if (c = #0) then raise Exception.create('Unexpected end of file!');
    if (length(s) >= 254) then raise Exception.create('Tag name too long!');
    s += c;
    next;
    goto xmlTagName;

    attributeName:
    if (c = '>') then begin
      if (el.selfClosing) then begin
        el := el.parentElement;
        if (el = nil) then raise Exception.Create('More elements where closed than created!');
      end;
      next;
      goto init;
    end else if (c = '=') or (c = '"') then begin
      if (c = '=') and (s <> '') then begin
        next;
        if (c <> '"') then raise Exception.Create('" expected, but ' + c + ' found!');
        t := '';
        next;
        goto attributeValue;
      end else raise Exception.Create('Unexpected character found!');
    end else if (c = #0) then raise Exception.create('Unexpected end of file!');
    if (length(s) >= 254) then raise Exception.create('Attribute name too long!');
    s += c;
    if (c = ' ') or (c = #10) or (c = #9) then s := ''; // attribute without value
    next;
    goto attributeName;

    attributeValue:
    if (c = '"') then begin
      el.setAttribute(s, XML.HTMLToUTF8(t));
      s := '';
      next;
      goto attributeName;
    end else if (c = #0) then raise Exception.create('Unexpected end of file!');
    t += c;
    next;
    goto attributeValue;

    xmlEnd:
    if (c = '>') then begin
      if (el = nil) then raise Exception.Create('End-Tag "' + s + '" found, which belongs to no element!');
      if (el.tagName <> UpperCase(s)) then raise Exception.Create('End-Tag "</' + el.tagName + '>" expected, but "</' + s + '>" found!');
      pn := el.childNodes.n;
      k := Length(pn)-1;
      if (k >= 0) and (pn[k].nodeType = ELEMENT_NODE) then begin
        if (TElement(pn[k]).textBreak and 8 = 8) then el.textBreak := el.textBreak or 4;
        TElement(pn[k]).textBreak := TElement(pn[k]).textBreak and (not 8);
      end else if (k >= 0) and (pn[k].nodeType = COMMENT_NODE) then begin
        if (TComment(pn[k]).textBreak and 8 = 8) then el.textBreak := el.textBreak or 4;
        TComment(pn[k]).textBreak := TComment(pn[k]).textBreak and (not 8);
      end;
      el := el.parentElement;
      if (el = nil) then raise Exception.Create('More elements where closed than created!');
      next;
      goto init;
    end else if (c = #0) then raise Exception.create('Unexpected end of file!');
    if (c <> ' ') and (c <> #10) and (c <> #9) then s += c;
    next;
    goto xmlEnd;

    commentNode:
    if (c = '>') then begin
      next;
      goto init;
    end else if (c = #0) then raise Exception.create('Unexpected end of file!');
    k := co.text.Count-1;
    if (c = #10) then co.text.add('')
    else co.text[k] := co.text[k] + c;
    next;
    goto commentNode;

    binaryBlock:
    next;
    if (c <> ' ') then raise Exception.create('Space expected but '+c+' found!');
    s := '';
    repeat begin
      next;
      if (length(s) >= 254) then raise Exception.create('Binary block size too large!');
      if (c = ' ') then break;
      s += c;
    end until false;
    try
      k := StrToInt(s);
    except
      raise Exception.create('Binary block size is an invalid integer!');
    end;
    if (pos + k > length(str)) then raise Exception.create('Size of binary block exceeds file size!');
    TCData(co) := TCData.create;
    setLength(TCData(co).data, k);
    move(str[pos], TCData(co).data[0], k);
    pos += k;
    next;
    if (c <> '>') then raise Exception.create('> expected but '+c+' found!');
    el.appendChild(co);
    next;
    co := nil;
    goto init;

    endOfFile:
    if (el <> Result) then raise Exception.create('End of file reached before all tags were closed!');
  except
    on E:Exception do begin
      Result.destroy;
      if (row >= 0) then raise Exception.create('Row '+inttostr(row)+', column '+inttostr(column)+': '+E.Message)
      else Exception.create('Position '+inttostr(pos)+': '+E.Message);
    end;
  end;
end;

function XML.TDocument.toStrings: TStrings;
var
  i : integer;
function doShift(shift : Integer) : AnsiString; inline;
var i : integer;
begin
 SetLength(Result, shift);
 for i := 1 to shift do Result[i] := ' '; // FillChar(Result[1], shift, ' ');
end;
procedure NodeToStr(n : TNode; shift : Integer);
var
  s : String;
  j : integer;
procedure ElToStr(n : TElement; shift : Integer);
var
  i,j : integer;
  s   : AnsiString;
begin
 s := '<' + n.tagName;
 for i := 0 to n.attributes.count - 1 do if (n.attributes[i].name <> '') then s := s + ' ' + n.attributes[i].name + '="' + XML.UTF8ToHTML(n.attributes[i].value, true) + '"';
 s := s + '>';
 n.shift := shift;
 if (n.textBreak and 1 = 1) or (Result.Count = 0) then Result.add(doShift(shift) + s) else Result[Result.Count-1] := Result[Result.Count-1] + s;
 if (n.textBreak and 2 = 2) then Result.add(doShift(shift+1));
 if (not n.selfClosing) then begin
  j := shift;
  if (n.textBreak and 2 = 2) or (n.textBreak and 4 = 4) then j := j + 1;
  for i := 0 to Length(n.childNodes.n) - 1 do NodeToStr(n.childNodes.n[i], j);
  s := '</' + n.tagName + '>';
  if (n.textBreak and 4 = 4) then Result.add(doShift(shift) + s) else Result[Result.Count-1] := Result[Result.Count-1] + s;
 end;
 if (n.textBreak and 8 = 8) then Result.add(doShift(shift));
end;
begin
 case n._nodeType of
  // UNKNOWN_NODEs are invisible in the String-Output
  ELEMENT_NODE : ElToStr(TElement(n), shift);
  TEXT_NODE : for j := 0 to TText(n).text.Count - 1 do
   if (j = 0) and (Result.Count > 0) then Result.Strings[Result.Count-1] := Result.Strings[Result.Count-1] + TText(n).text[j]
   else Result.add(TText(n).text[j]);
  COMMENT_NODE : if (n is TCData) then begin
   s := '<!BINDATA '+inttostr(length(TCData(n).data))+' ';
   j := length(s)+1;
   setLength(s, length(s)+length(TCData(n).data)+1);
   s[length(s)] := '>';
   move(TCDAta(n).data[0], s[j], length(TCDAta(n).data));
   Result.add(s);
  end else for j := 0 to TComment(n).text.Count - 1 do begin
   if (j = 0) then s := '<!' + TComment(n).text[j]
   else s := TComment(n).text[j];
   if (j = TComment(n).text.Count - 1) then s := s + '>';
   if (TComment(n).textBreak and 1 = 1) or (Result.Count = 0) then Result.add(doShift(shift) + s) else Result[Result.Count-1] := Result[Result.Count-1] + s;
   if (TComment(n).textBreak and 8 = 8) and (j = TComment(n).text.Count - 1) then Result.add(doShift(shift));
  end;
 end;
end;
begin
 Result := TStringList.Create;
 for i := 0 to childNodes.count - 1 do NodeToStr(childNodes[i], 0);
end;


initialization
XML.dtd := nil;
XML.symbolTable := nil;

if (XML.HTMLToUTF8('hallo 123 &#34; &auml;&ouml;&uuml;&szlig; &#x263B; &Auml;&darr;&dArr; &bla;bla &blablablablabla&#9787; ;; &#x;') <> 'hallo 123 " äöüß ☻ Ä↓⇓ &bla;bla &blablablablabla☻ ;; &#x;') then raise Exception.Create('HTMLToUTF8 give wrong result!');

end.
